#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <sys/resource.h>

static void msg (const char * fmt, ...) {
  fputs ("[crack] ", stdout);
  va_list ap;
  va_start (ap, fmt);
  vprintf (fmt, ap);
  va_end (ap);
  fputc ('\n', stdout);
  fflush (stdout);
}

static void die (const char * fmt, ...) {
  fputs ("*** 'crack' error: ", stderr);
  va_list ap;
  va_start (ap, fmt);
  vfprintf (stderr, fmt, ap);
  va_end (ap);
  fputc ('\n', stderr);
  exit (1);
}

static const char * usage =
"crack [ -h ] [ -b <bits> ] [ -s <key> ] [ -p <plain> ]\n"
;

static int bits = 32;
static unsigned long key;
static unsigned long mask;
static unsigned long plain;
static unsigned long cypher;
static unsigned long encryptions;

static unsigned long str2word (const char * str) {
  unsigned long res = 0;
  const char * p;
  for (p = str; *p; p++) {
    unsigned long tmp = res & (0xful << 60);
    res <<= 4;
    res += *p;
    if (tmp) res ^= tmp >> 60;
  }
  return res;
}

static unsigned long encrypt (unsigned long k, unsigned long d) {
  encryptions++;
  return (k * d) & mask;
}

static double wall_clock_time () {
  double res = 0;
  struct timeval tv;
  if (!gettimeofday (&tv, 0))
  res = 1e-6 * tv.tv_usec, res += tv.tv_sec;
  return res;
}

static double process_time () {
  struct rusage u;
  if (getrusage (RUSAGE_SELF, &u)) return 0;
  double res = u.ru_utime.tv_sec + 1e-6 * u.ru_utime.tv_usec;
  res += u.ru_stime.tv_sec + 1e-6 * u.ru_stime.tv_usec;
  return res;
}

int main (int argc, char ** argv) {
  for (int i = 1; i < argc; i++) {
    if (!strcmp (argv[i], "-h"))
      fputs (usage, stdout), exit (0);
    else if (!strcmp (argv[i], "-b")) {
      if (++i == argc) die ("argument to '-b' missing (try '-h')");
      if ((bits = atoi (argv[i])) < 1 || bits > 64)
	die ("invalid bits argument in '-b %s'", argv[i]);
    } else if (!strcmp (argv[i], "-k")) {
      if (++i == argc) die ("argument to '-k' missing (try '-h')");
      key = str2word (argv[i]);
    } else if (!strcmp (argv[i], "-p")) {
      if (++i == argc) die ("argument to '-p' missing (try '-h')");
      plain = str2word (argv[i]);
    } else die ("invalid command line option '%s' (try '-h')", argv[i]);
  }
  if (!(key & 1)) key++;
  msg ("using %d bits", bits);
  mask = (1ul << bits) - 1;
  msg ("mask        0x%016lx", mask);
  key &= mask;
  plain &= mask;
  msg ("key word    0x%016lx", key);
  msg ("plain word  0x%016lx", plain);
  cypher = encrypt (key, plain);
  msg ("cypher word 0x%016lx", cypher);

  // Simulate a known plain text attack, e.g., pretend not to know the key
  // and search for it.

  double started = wall_clock_time ();

  unsigned long k = 1;
  int found = 0;
  for (;;) {
    unsigned long tmp = encrypt (k, plain);
    if (tmp == cypher) found = 1;
    if (found || k == mask) break;
    k += 2;
  }
  double p = process_time ();
  double w = wall_clock_time () - started;
  if (found) msg ("found key   0x%016lx", k);
  else msg ("did not find key!");
  msg ("after %ul encryptions", encryptions);
  msg ("%.2f seconds process time", p);
  msg ("%.2f seconds wall-clock", w);
  return !found;
}
